#!/bin/sh
# the next line restarts using wish \
exec wish "$0" "$@"

#############################################################################
# PgcEdit.tcl Copyright 2004 r0lZ
# Program main entry.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

#############################################################################
#
# PgcEdit requires Tcl/Tk 8.4 or greather
#
#############################################################################

set ::program_version 0.3.8
set ::config(debug) 0


wm withdraw .
# necessary for freeWrap 5.61
if {!$tcl_interactive} {
	catch {console hide}
}
update idletasks

#catch {encoding system cp1252}
#catch {encoding system utf-8}
catch {encoding system identity}

if {$tcl_version < 8.4} {
	tk_messageBox -icon error -title "PgcEdit $::program_version error" -parent . \
		-message "PgcEdit requires Tcl/Tk version 8.4 or greather.\nYou are currently running Tcl/Tk version $tcl_patchLevel.\nSorry."
	exit 1000
}

#############################################################################
## Debug print.
## Usage: set ::config(debug) <triggerlevel>
##        dprint <level> <message>

proc dprint {l t} {
	catch {if {$::config(debug)>=$l} {puts $t}}
}

#############################################################################
## show_msg
## Show message in the status line
proc show_msg {msg} {
	if [winfo exists .mf.st.status] {
		.mf.st.status configure -text [string map {"\n" "  "} $msg]
		update idletasks
	}
}

#############################################################################
## show_warning
## Show message in a message box and in the status line.
## Usage: show_warning -message "message" [-title "title"] [-icon info|warning|error] [-parent window]
proc show_warning {args} {
	set title "PgcEdit"
	set icon "info"
	set parent .

	set idx 0
	set opt [lindex $args $idx]
	while {$opt != ""} {
		incr idx
		set val [lindex $args $idx]
		incr idx
		switch -- $opt {
			"-message" {set msg $val}
			"-title" {set title $val}
			"-icon" {set icon $val}
			"-parent" {set parent $val}
		}
		set opt [lindex $args $idx]
	}
	show_msg $msg
	if {$msg != ""} {
		if { ! $::config(expert_mode) || $icon == "error"} {
			tk_messageBox -message $msg -title $title -icon $icon -parent .
		}
	}
}

proc tk_chooseNewDirectory {args} {
	if {$::config(use_file_browser)} {
		set destfile [eval tk_getSaveFile -initialfile "VIDEO_TS.IFO" \
				-filetypes \{\{"All files" \{*.*\}\} \{"DVD files" \{*.IFO *.BUP *.VOB\}\} \{"VIDEO_TS.IFO" \{VIDEO_TS.IFO\}\}\} \
				$args]
		if {$destfile == ""} {return ""}
		set destdir [file dirname $destfile]
	} else {
		set destdir [eval tk_chooseDirectory $args -mustexist false]
		if {$destdir == ""} {return ""}
	}

	if {! [file exists $destdir]} {
		set pathname $destdir
		while {![file exists $pathname]} {
			set newdir [file dirname $pathname]
			if {$pathname == $newdir} {
				set pathname ""
			} else {
				set pathname $newdir
			}
		}
		set lastpart [file tail $pathname]
		set newparts [string range $destdir [expr [string length $pathname]+1] end]
		set newpart1 [lindex [file split $newparts] 0]
		if {$lastpart == $newpart1} {
			set newpartslist [file split $newparts]
			for {set i 1} {$i < [llength $newpartslist]} {incr i} {
				set pathname [file join $pathname [lindex $newpartslist $i]]
			}
		} else {
			set pathname $destdir
		}

		set r [tk_messageBox -title "Browse for Folder" -icon question \
				-message "Folder \"[file nativename $pathname]\\\" does not exist.\nCreate ?" \
				-parent . -type yesno]
		if {$r == "no"} {
			return ""
		} else {
			set err [catch {file mkdir $pathname} errmsg]
			if {$err!=0} {
				show_warning -title "PgcEdit: Save IFOs as" -icon error \
						-message "Could not create folder \"$pathname\":\n$errmsg" \
						-parent .
				return ""
			}
		}
	} else {
		set pathname $destdir
	}
	return $pathname
}

proc tk_chooseOldDirectory {args} {
	if {$::config(use_file_browser)} {
		set destfile [eval tk_getOpenFile -initialfile "VIDEO_TS.IFO" \
				-filetypes \{\{"All files" \{*.*\}\} \{"DVD files" \{*.IFO *.BUP *.VOB\}\} \{"VIDEO_TS.IFO" \{VIDEO_TS.IFO\}\}\} \
				-defaultextension 2 $args]
		if {$destfile == ""} {return ""}
		return [file dirname $destfile]
	} else {
		set destdir [eval tk_chooseDirectory $args -mustexist true]
		return $destdir
	}
}

#############################################################################
## Procedure:  init_user_prefs
## Init user modifiables constants
namespace eval ::tr {}

proc init_user_prefs {} {
	global argv0
	# New to v0.3: config files are stored in "~/Application Data/PgcEdit/"
	# instead of "~/", under Windows.
	set ::config(config_directory) $::env(HOME)
	if [file exists [file join $::env(HOME) "Application Data"]] {
		set ::config(config_directory) [file join $::env(HOME) "Application Data/PgcEdit"]
		if {![file exists $::config(config_directory)]} {
			file mkdir $::config(config_directory)
			catch {file rename -force [file join $::env(HOME) ".pgcedit.cfg"]   $::config(config_directory)}
			catch {file rename -force [file join $::env(HOME) ".pgcedit.tools"] $::config(config_directory)}
		}
	}

	set ::config(restore_defaults) false

	if {$::tcl_platform(platform) == "windows"} {
		set rootdir "C:"
		foreach i {C: D: E: F: G: H: I: J: K: L: M: N: O: P: Q: R: S: T: U: V: W: X: Y: Z:} {
			if [file exists $i] {
				set rootdir "$i\\"
			}
		}
		set progfiles "${rootdir}Program files\\"
	} else {
		set rootdir {/}
		set progfiles {/}
	}

	# Automatically save setups on program exit?
	set ::config(save_setups_on_exit) true

	# fonts
	label .testfont
	set ::config(global_font) [.testfont cget -font]
	destroy .testfont
	if {$::tcl_platform(platform) == "windows"} {
		set ::config(fixed_width_font) {{Courier} 9}
	} else {
		set ::config(fixed_width_font) {{Courier} 8}
		set ::config(global_font) [list [lindex $::config(global_font) 0] 8]
	}

	# Tearoff menus
	if {$::tcl_platform(platform) == "unix" && [string first "MAC" [string toupper $::tcl_platform(machine)]] >= 0} {
		set ::config(tearoff_menus) false
	} else {
		set ::config(tearoff_menus) true
	}

	# Automatically launch Open DVD dialog at startup?
	set ::config(auto_open_new_dvd) false

	# Use file browser instead of folder browser to Open/Save DVD?
	set ::config(use_file_browser) false

	# Automatically create backup when a DVD is opened?
	set ::config(auto_create_backup) true

	# Show the PGC selector at startup?
	set ::config(selector) true

	# which side to display the selector?
	set ::config(selector_side) left

	# height of the selector (for top and bottom sides only)?
	set ::config(selector_height) 8

	# Show the action buttons at startup?
	set ::config(buttons) true

	# Show the toolbar at startup?
	set ::config(toolbar) true

	# Automatically show the action buttons at startup?
	set ::config(searchstring) ""

	# Automatically adjust Command Editor window size?
	set ::config(adjust_win_size) false

	# Renumber GOTSs?
	set ::config(renumber_gotos) false

	# Highlight errors and Goto targets?
	set ::config(highlight_gotos) true

	# Display [hex codes] in Info -> Dumps
	set ::config(dump_hex_codes) true

	# Expert mode?
	set ::config(expert_mode) false

	# The startup directory for the DVD file browser:
	set ::config(initialdir) $rootdir

	# The startup directory for the Load Menu file browser:
	set ::config(importmenuinitialdir) $rootdir

	# The startup directory for the scratchbook file browser:
	set ::config(scratchbook_initialfile) $rootdir

	# The startup directory for the scratchbook file browser:
	set ::config(import_initialdir) $rootdir

	# The startup filename for the dump file browser:
	set ::config(dumpinitialfile) "${rootdir}dvd_dump.txt"

	# Size and position of the main window
	set ::config(main_win_geometry) 700x400

	# Position of the Command Editor window
	set ::config(ced_win_geometry) ""

	# Position of the Trace Info window
	set ::config(trace_info_win_geometry) ""

	# Position of the Menu Buttons Info window
	set ::config(menubuttons_win_geometry) ""

	# Divisor for the Menu Buttons Video Screen window
	set ::config(menubuttons_screen_divisor) 1.333333

	# Trace Info View flags
	set ::tr::view_gprms 1
	set ::tr::view_sprms 1
	set ::tr::view_puos  0
	set ::tr::view_input 1
	set ::tr::view_slabs 1

	# Show libdvdread warnings when parsing menu buttons ?
	set ::config(show_libdvdread_warnings) false

	# v0.1.3 bug in config file workaround
	set ::config(selector_win_geometry) ""

	# build default scratchbook filename
	set ::config(default_scratchbook_filename) \
			[file join $::config(config_directory) .pgcedit.scratchbook.sbk]

	# Overwrite default values with user defined ones
	set configfile [file join $::config(config_directory) ".pgcedit.cfg"]
	if [file exists $configfile] {
		if {[catch {source $configfile} err] != 0} {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}

	# if ::config(initialdir) does not exists, find the first directory that exists
	while {![file exists $::config(initialdir)]} {
		set newdir [file dirname $::config(initialdir)]
		if {$::config(initialdir) == $newdir} {
			set ::config(initialdir) $rootdir
		} else {
			set ::config(initialdir) $newdir
		}
	}

	# if ::config(importmenuinitialdir) does not exists, find the first directory that exists
	while {![file exists $::config(importmenuinitialdir)]} {
		set newdir [file dirname $::config(importmenuinitialdir)]
		if {$::config(importmenuinitialdir) == $newdir} {
			set ::config(importmenuinitialdir) $rootdir
		} else {
			set ::config(importmenuinitialdir) $newdir
		}
	}

	# if ::config(dumpinitialfile) does not exists, find the first directory that exists
	set pathname [file dirname $::config(dumpinitialfile)]
	set filename [file tail    $::config(dumpinitialfile)]
	while {![file exists $pathname]} {
		set newdir [file dirname $pathname]
		if {$pathname == $newdir} {
			set pathname $rootdir
		} else {
			set pathname $newdir
		}
	}
	set ::config(dumpinitialfile) [file join $pathname $filename]

	# Tools
	set ::config(tools) {}

	# Overwrite default values with user defined ones
	set configfile [file join $::config(config_directory) ".pgcedit.tools"]
	if [file exists $configfile] {
		if {[catch {source $configfile} err] != 0} {
			show_warning -icon error -title "PgcEdit $::program_version error" \
				-message "Error in config file $configfile:\n$err" -parent .
		}
	}

	# set the "associate PgcEdit with .XXX" menu flags
	# based upon current registry associations
	# Only used if PgcEdit is freeWrapped and run under Windows
	# The catchs are supposed to avoid a bug in ::freewrap::shell_getCmd_imp
	if {$::tcl_platform(platform) == "windows" && [info exists ::freewrap::progname]} {
		if [catch {set ::config(shortcut_startmenu) \
				[::winshellext::is_shortcut_to_PgcEdit startmenu]}] {
			set ::config(shortcut_startmenu) 0
		}
		if [catch {set ::config(shortcut_sendto) \
				[::winshellext::is_shortcut_to_PgcEdit sendto]}] {
			set ::config(shortcut_sendto) 0
		}
		if [catch {set ::config(shortcut_desktop) \
				[::winshellext::is_shortcut_to_PgcEdit desktop]}] {
			set ::config(shortcut_desktop) 0
		}
		if [catch {set ::config(assoicated_with_IFO) \
				[::winshellext::shell_is_associated ".IFO" "PgcEdit.exe"]}] {
			set ::config(assoicated_with_IFO) 0
		}
		if [catch {set ::config(assoicated_with_BUP) \
				[::winshellext::shell_is_associated ".BUP" "PgcEdit.exe"]}] {
			set ::config(assoicated_with_BUP) 0
		}
		if [catch {set ::config(web_browser) \
				[lindex [string map {\\ /} [::freewrap::shell_getCmd_imp .html open]] 0] }] {
			set ::config(web_browser) "C:/Program Files/Internet Explorer/iexplore.exe"
		}
		set ::config(min_bd) 0
	} else {
		if {$::tcl_platform(platform) == "windows"} {
			set ::config(web_browser) "C:/Program Files/Internet Explorer/iexplore.exe"
		} else {
			set ::config(web_browser) ""
		}
		set ::config(min_bd) 2
	}
}

#############################################################################
## Procedure:  save_user_prefs
## save setups
proc save_user_prefs {} {
	set txt "# PgcEdit config file.  Saved automatically.  Do not edit!\n"
	set txt "${txt}\nset ::config(global_font) [list $::config(global_font)]"
	set txt "${txt}\nset ::config(fixed_width_font) [list $::config(fixed_width_font)]"
	set txt "${txt}\nset ::config(tearoff_menus) $::config(tearoff_menus)"
	set txt "${txt}\nset ::config(save_setups_on_exit) $::config(save_setups_on_exit)"
	set txt "${txt}\nset ::config(use_file_browser) $::config(use_file_browser)"
	set txt "${txt}\nset ::config(auto_open_new_dvd) $::config(auto_open_new_dvd)"
	set txt "${txt}\nset ::config(auto_create_backup) $::config(auto_create_backup)"
	set txt "${txt}\nset ::config(selector) $::config(selector)"
	set txt "${txt}\nset ::config(selector_side) $::config(selector_side)"
	set txt "${txt}\nset ::config(selector_height) $::config(selector_height)"
	set txt "${txt}\nset ::config(buttons) $::config(buttons)"
	set txt "${txt}\nset ::config(toolbar) $::config(toolbar)"
	set txt "${txt}\nset ::config(searchstring) [list $::config(searchstring)]"
	set txt "${txt}\nset ::config(adjust_win_size) $::config(adjust_win_size)"
	set txt "${txt}\nset ::config(highlight_gotos) $::config(highlight_gotos)"
	set txt "${txt}\nset ::config(dump_hex_codes) $::config(dump_hex_codes)"
	set txt "${txt}\nset ::config(renumber_gotos) $::config(renumber_gotos)"
	set txt "${txt}\nset ::config(expert_mode) $::config(expert_mode)"
	set txt "${txt}\nset ::config(initialdir) [list [file nativename $::config(initialdir)]]"
	set txt "${txt}\nset ::config(importmenuinitialdir) [list [file nativename $::config(importmenuinitialdir)]]"
	set txt "${txt}\nset ::config(scratchbook_initialfile) [list [file nativename $::config(scratchbook_initialfile)]]"
	set txt "${txt}\nset ::config(import_initialdir) [list [file nativename $::config(import_initialdir)]]"
	set txt "${txt}\nset ::config(dumpinitialfile) [list [file nativename $::config(dumpinitialfile)]]"
	if {[winfo width .]==[winfo screenwidth .] && [winfo rootx .]==0} {
		set txt "${txt}\nset ::config(main_win_geometry) $::config(main_win_geometry)"
	} else {
		set txt "${txt}\nset ::config(main_win_geometry) [wm geometry .]"
	}
	if [winfo exists .cmdeditor] {
		set txt "${txt}\nset ::config(ced_win_geometry) [wm geometry .cmdeditor]"
	} else {
		set txt "${txt}\nset ::config(ced_win_geometry) \"$::config(ced_win_geometry)\""
	}
	if [winfo exists .trinfo] {
		set txt "${txt}\nset ::config(trace_info_win_geometry) [wm geometry .trinfo]"
	} else {
		set txt "${txt}\nset ::config(trace_info_win_geometry) \"$::config(trace_info_win_geometry)\""
	}

	if [info exists ::menubuttons::win_geometry] {
		set txt "${txt}\nset ::config(menubuttons_win_geometry) \"$::menubuttons::win_geometry\""
	} else {
		set txt "${txt}\nset ::config(menubuttons_win_geometry) \"$::config(menubuttons_win_geometry)\""
	}
	set txt "${txt}\nset ::config(menubuttons_screen_divisor) $::config(menubuttons_screen_divisor)"
	set txt "${txt}\nset ::tr::view_gprms $::tr::view_gprms"
	set txt "${txt}\nset ::tr::view_sprms $::tr::view_sprms"
	set txt "${txt}\nset ::tr::view_puos  $::tr::view_puos"
	set txt "${txt}\nset ::tr::view_input $::tr::view_input"
	set txt "${txt}\nset ::tr::view_slabs $::tr::view_slabs"

	if {$::config(debug) > 0} {
		set txt "${txt}\nset ::config(debug) $::config(debug)"
	}
	set txt "${txt}\nset ::config(show_libdvdread_warnings) $::config(show_libdvdread_warnings)"
	set txt "${txt}\n"

	saveFile [file join $::config(config_directory) ".pgcedit.cfg"] auto $txt
}

#############################################################################
## Procedure:  cleanexit
## save setups and exit
proc cleanexit {code} {
	if {$::dvd_modified && $code == 0 && !$::config(expert_mode)} {
		set r [tk_messageBox -message "The DVD has been modified.\n\nOK to quit?" \
				-icon warning -title "PgcEdit" -type yesno -default no]
		if {$r == "no"} return
	}
	#bind . <Destroy> {}
	if {$::config(save_setups_on_exit) && $::config(restore_defaults)==false} {
		save_user_prefs
	}
	if $::config(debug) {
		tk_messageBox -icon info -message "Exiting (code=$code)" -title "PgcEdit"
	}
	exit $code
}

#############################################################################
## Procedure:  errmsg
## Display an error message
proc errmsg {errnum msg1 msg2} {
	# Display a message
	puts "Error $errnum: $msg1"
	puts $msg2
	if [info exists errorInfo] {catch {puts $errorInfo}}
	if {$::config(debug) && $::tcl_platform(platform) == "windows"} {
		console show
	}
	show_warning -icon error -title "PgcEdit $::program_version error" \
		-message "Error $errnum $msg1\n$msg2" -parent .
}

#############################################################################
## Procedure:  readFile
## read a file and return the datas
## return "" in case of error
proc readFile {filename translation} {
	set errnum [catch {set in [open $filename r]} err]
	if {$errnum} {
		errmsg $errnum "opening file for read" $err
		return ""
	}
	fconfigure $in -translation $translation
	set errnum [catch {set data [read $in]} err]
	if {$errnum} {
		close $in
		errmsg $errnum "reading file" $err
		return ""
	}
	close $in
	return $data
}

#############################################################################
## Procedure:  saveFile
## save datas in a file
## return error number, or 0 for OK
## if $args == "-append", then append to end of an existing file, else overwrite
proc saveFile {filename translation data args} {
	if {$args == "-append"} {
		set mode a
		set modetext "append"
	} else {
		set mode w
		set modetext "write"
	}
	set errnum [catch {set out [open $filename $mode]} err]
	if {$errnum} {
		errmsg $errnum "opening file for $modetext" $err
		return $errnum
	}
	fconfigure $out -translation $translation
	if {$translation == "binary"} {
		set errnum [catch {puts -nonewline $out $data} err]
	} else {
		set errnum [catch {puts $out $data} err]
	}
	if {$errnum} {
		errmsg $errnum "writing file" $err
	}
	close $out
	return $errnum
}

#############################################################################
## Procedure:  b2x
## Convert a binary string to hexadecimal
proc b2x {binarystring} {
	binary scan $binarystring H* hex
	return $hex
}

#############################################################################
## Procedure:  b2i
## Convert a binary string to decimal
proc b2i {binarystring} {
	binary scan $binarystring H* hex
	return [expr 0x$hex]
}

#############################################################################
## Procedure:  d2c
## Convert a value to binary characters
proc d2c {val numbytes} {
	switch $numbytes {
		1 {return [binary format c $val]}
		2 {return [binary format S $val]}
		4 {return [binary format I $val]}
		8 {return [binary format W $val]}
	}
}

#############################################################################
## Procedure:  get_int
## Get an integer from inside the binary data
proc get_int {binarystring_upvar start_offset length} {
	upvar $binarystring_upvar binarystring
	set s [string range $binarystring $start_offset [expr $start_offset + $length - 1]]
	binary scan $s H* hex
	if {$hex==""} {
		set txt "Fatal error!\n"
		set txt "$txt\nPgcEdit is trying to read data after the"
		set txt "$txt\nend of a table."
		set txt "$txt\nThis means that a Length Of Table value"
		set txt "$txt\nis probably wrong.\n"
		set txt "$txt\nPlease try to fix the problem with IfoEdit.\n"
		set txt "$txt\nThe missing value may be fixed to 0, but"
		set txt "$txt\nthis could lead to strange results.\n"
		set txt "$txt\nDo you want to fix the value and continue"
		set txt "$txt\nanyway? (NOT RECOMMENDED!)"
		set r [tk_messageBox -title "PgcEdit fatal error" -icon error -message $txt \
				-parent . -type yesno -default yes]
		if {$r == "no"} {error "Attempt to read data after end of table"}
		set hex 00
	}
	return [expr 0X$hex]
}

#############################################################################
## Procedure:  set_int
## Set an integer inside the binary data whose variable is passed in arguments
proc set_int {binary_upvar start_offset value length} {
	upvar $binary_upvar binarystring
	set binarystring [string replace $binarystring $start_offset \
			[expr $start_offset + $length - 1] [d2c $value $length]]
	return
}

#############################################################################
# return the number of pre, post or cell commands in the provided commands
proc num_type_cmds {type cmds_upvar} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {return [expr [lindex $idxs 1]-[lindex $idxs 0]-1]}
		post {return [expr [lindex $idxs 2]-[lindex $idxs 1]-1]}
		cell {return [expr [llength $cmds]-[lindex $idxs 2]-1]}
	}
}

#############################################################################
# return the pre, post or cell commands from the provided commands
proc get_type_cmds {type cmds_upvar} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {return [lrange $cmds [expr [lindex $idxs 0]+1] [expr [lindex $idxs 1]-1]]}
		post {return [lrange $cmds [expr [lindex $idxs 1]+1] [expr [lindex $idxs 2]-1]]}
		cell {return [lrange $cmds [expr [lindex $idxs 2]+1] end]}
	}
}

#############################################################################
# return the type ("pre", "post" or "cell") of the Nth. command in the provided commands.
# If the cmd number is a pre, post or cell label, return "* pre", "* post" or "* cell".
proc get_cmd_type {cmds_upvar index} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	if {$index==[lindex $idxs 0]} {return "* pre"}
	if {$index==[lindex $idxs 1]} {return "* post"}
	if {$index==[lindex $idxs 2]} {return "* cell"}
	if {$index<[lindex $idxs 1]} {return "pre"}
	if {$index<[lindex $idxs 2]} {return "post"}
	return "cell"
}

#############################################################################
# return the line number of the command into his type section.
# If the cmd number is a pre, post or cell label, return 0.
proc get_cmd_linenum {cmds_upvar index} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	if {$index<[lindex $idxs 1]} {return $index}
	if {$index<[lindex $idxs 2]} {return [expr $index - [lindex $idxs 1]]}
	return [expr $index - [lindex $idxs 2]]
}

#############################################################################
# replace the pre, post or cell commands in the commands list by the provided commands
proc set_type_cmds {type cmds_upvar newcmds} {
	upvar $cmds_upvar cmds

	set idxs [lsearch -all $cmds {[*]*}]
	switch $type {
		pre  {
			set idx1 [lindex $idxs 0]
			set idx2 [lindex $idxs 1]
		}
		post {
			set idx1 [lindex $idxs 1]
			set idx2 [lindex $idxs 2]
		}
		cell {
			set idx1 [lindex $idxs 2]
			set idx2 [llength $cmds]
		}
	}
	set cmds [concat [lrange $cmds 0 $idx1] $newcmds [lrange $cmds $idx2 end]]
}

#############################################################################
# refresh the listbox with the current PGC
proc refresh_current_pgc {} {
	if $::scratchbook_active {
		return
	}
	set selectionindexes [::main::get_selection_indexes]
	set type [lindex $::editor_cmds_from 0]
	set pgcindex [split [lindex $::editor_cmds_from 1] ","]
	set vts [lindex $pgcindex 0]
	switch $type {
		pgc {
			::main::display_pgc $type $vts [lindex $pgcindex 1]
		}
		menu {
			::main::display_pgc $type $vts [lindex $pgcindex 1] [lindex $pgcindex 2]
		}
	}
	::main::set_listbox_selection [lindex $selectionindexes 0] [lindex $selectionindexes 1]
}

#############################################################################
## PROGRAM MAIN ENTRY
#############################################################################

# Parse arguments
if {$argc == 1 && ([lindex $argv 0] == "/?" || [lindex $argv 0] == "-h")} {
	puts "PgcEdit v $::program_version"
	puts "Command line usage:"
	puts "> PgcEdit \[\"DVD pathname\"\]"
	puts "or"
	puts "> PgcEdit \[\"DVD pathname\any file\"\]"
	puts "The second form is used to associate PgcEdit with IFO (or BUP) files."
	exit 0
}

set ::config(tempdir) "NOT_ASSIGNED"
catch {set ::config(tempdir) $::env(TEMP)}
if {! [file isdirectory $::config(tempdir)]} {
	catch {set ::config(tempdir) $::env(TMP)}
}
if {! [file isdirectory $::config(tempdir)]} {
	if {[file isdirectory "/tmp"] && [file writable "/tmp"]} {
		catch {set ::config(tempdir) "/tmp"}
	}
}

# Full path is needed for FreeWrap
set scriptname [info script]
if [info exists ::freewrap::progname] {
	# FreeWrap as wish or Wrapped script
	if {$scriptname == ""} {
		# Wrapped script
		set config(apppath) "/Tcl/work/PGCEDIT"
	} else {
		# FreeWrap as wish
		set config(apppath) [file dirname $scriptname]
	}
	source "$config(apppath)/lib/winshellext.tcl"
	set ::config(bindir) [file join [file dirname $argv0] "bin"]
	if {! [file exists $::config(tempdir)]} {
		file mkdir [file join [file dirname $argv0] "temp"]
		set ::config(tempdir) [file join [file dirname $argv0] "temp"]
	}

	set copy 0
	if {! [file exists [file join $::config(bindir) "binversion.tcl"]]} {
		set copy 1
	} else {
		source [file join $::config(bindir) "binversion.tcl"]
		if {$bin_version != $::program_version} {
			set copy 1
		}
	}
	if $copy {
		file mkdir $::config(bindir)
		if {$::tcl_platform(platform) == "windows"} {
			saveFile [file join $::config(bindir) "button_dump.exe"] binary [readFile [file join $config(apppath) "bin/button_dump.exe"] binary]
			catch {file delete [file join $::config(bindir) "BlankCell.exe"]}	;# not needed now
		}
		if {$::tcl_platform(platform) == "unix"} {
			if {[string first "MAC" [string toupper $::tcl_platform(machine)]] >= 0} {
				saveFile [file join $::config(bindir) "button_dump_osx"] binary [readFile [file join $config(apppath) "bin/button_dump_osx"] binary]
				catch [file attributes [file join $::config(bindir) "button_dump_osx"] -permissions u+rx,g+rx,o+rx]
			} else {
				saveFile [file join $::config(bindir) "button_dump_linux"] binary [readFile [file join $config(apppath) "bin/button_dump_linux"] binary]
				catch [file attributes [file join $::config(bindir) "button_dump_linux"] -permissions u+rx,g+rx,o+rx]
			}
		}
		if [file exists [file join $::config(bindir) "cygwin1.dll"]] {
			file delete [file join $::config(bindir) "cygwin1.dll"]
		}
		saveFile [file join $::config(bindir) "binversion.tcl"] auto "set bin_version $::program_version\n"
		puts "Binaries v$::program_version copied."
	}
} else {
	set config(apppath) [file dirname $scriptname]
	set ::config(bindir) [file join [file dirname $scriptname] "bin"]
	if {! [file exists $::config(tempdir)]} {
		set ::config(tempdir) "/tmp"
	}
	if {! [file exists $::config(tempdir)]} {
		file mkdir [file join [file dirname $scriptname] "temp"]
		set ::config(tempdir) [file join [file dirname $scriptname] "temp"]
	}

	if {$::tcl_platform(platform) != "macintosh"} {
		source [file join $::config(bindir) "binversion.tcl"]
		if {$bin_version != $::program_version} {
			tk_messageBox -icon error -title "PgcEdit $::program_version error" -parent . \
				-message "The binaries (in bin subdirectory)\nare either missing or obsolete!\nPlease reinstall PgcEdit."
		}
		if [file exists [file join $::config(bindir) "cygwin1.dll"]] {
			file delete [file join $::config(bindir) "cygwin1.dll"]
		}
	}
	if {$::tcl_platform(platform) == "unix"} {
		catch {file attributes [file join $::config(bindir) "button_dump_linux"] -permissions u+rx,g+rx,o+rx}
		catch {file attributes [file join $::config(bindir) "button_dump_osx"]   -permissions u+rx,g+rx,o+rx}
	}
}

# init user preferences
init_user_prefs

# disable the dprint routine if it is not needed, for speed purposes
if {$::config(debug) == 0} {
	proc dprint {args} {}
}

# init some system defaults
# we need to know some widget default colors
set config(systembuttonface) [[button .w] cget -background]; destroy .w
set config(disabledforeground) [[button .w] cget -disabledforeground]; destroy .w
if {$::tcl_platform(platform) == "windows"} {
	set config(systemhighlight) systemhighlight
} else {
	set config(systemhighlight) gray75
}

option add *font $::config(global_font)
option add *tearOff $::config(tearoff_menus)
option add *Scale*troughColor white
option add *Scale*sliderLength 15
option add *Scale*highlightThickness 1
option add *Scale*borderWidth 1
option add *Scale*orient horizontal
option add *Menubutton*relief raised
option add *Menubutton*indicatorOn true
option add *Menubutton*highlightThickness 1
option add *Menubutton*takeFocus _focus_menubutton
proc _focus_menubutton {w} {if {[$w cget -state]=="disabled"} {return 0} else {return 1}}

# read program modules
source $config(apppath)/lib/tk_menu_bugfix.tcl	;# MenuUnpost Tk bug fixed by r0lZ
source $config(apppath)/lib/readdvd.tcl
source $config(apppath)/lib/writedvd.tcl
source $config(apppath)/lib/disasm.tcl
source $config(apppath)/lib/utils.tcl
source $config(apppath)/lib/macros.tcl
source $config(apppath)/lib/toolscfg.tcl
source $config(apppath)/lib/cmdedit.tcl
source $config(apppath)/lib/pgceditor.tcl
source $config(apppath)/lib/srptinfo.tcl
source $config(apppath)/lib/streams.tcl
source $config(apppath)/lib/menubuttons.tcl
source $config(apppath)/lib/blankcell.tcl
source $config(apppath)/lib/toolbar.tcl
source $config(apppath)/lib/stringreq.tcl
source $config(apppath)/lib/fontreq.tcl
source $config(apppath)/lib/calculator.tcl
source $config(apppath)/lib/main.tcl			;# main program starts here

# for Tcl/Tk programmers: add your private functions in file ~/.pgcedit.user
set userfile [file join $::config(config_directory) "pgcedit.user.tcl"]
if [file exists $userfile] {source $userfile}


# initialize the command editor
::ced::cmdeditor
