#############################################################################
# streams.tcl Copyright 2004 r0lZ
# Compile informations about all Video/Audio/Sub-picture streams in the DVD.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

namespace eval ::streams {
namespace export get_streams_info
namespace import ::readdvd::*

set ::streams::audio_code_ext_list {
	{unspecified}
	{normal}
	{for visual impaired}
	{director comment}
	{alt director comment}
}
set ::streams::subpic_code_ext_list {
	{unspecified}
	{normal}
	{large}
	{children}
	-
	{normal captions}
	{large captions}
	{children captions}
	-
	{forced}
	-
	-
	-
	{director comments}
	{large director comments}
	{comments for children}
}
set ::streams::illegalextlabel "ILLEGAL CODE EXTENSION"

####### main proc
proc get_streams_info {} {
	if {! [info exists ::dvddir]} {
		bell
		return ""
	}

	set txt "DVD: \"$::dvddir/\""
	set txt "$txt\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set txt "$txt\nProvider ID: \"[::utils::get_provider_ID]\""
	set txt "$txt\nNumber of VTS: $::pgcs(numvts)"
	set txt "$txt\n"

	set numaudios [get_int ::raw_sectors(0,VMGM_MAT) 0X102 2]
	set numsubpics [get_int ::raw_sectors(0,VMGM_MAT) 0X154 2]
	set txt "$txt\nVMG Menu streams attributes (VIDEO_TS.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
	set txt "$txt\n- Video:    [video_description [get_int ::raw_sectors(0,VMGM_MAT) 0X100 2] true]"
	for {set s 0} {$s < $numaudios} {incr s} {
		set offset [expr 0X104 + ( $s * 8 )]
		set attr [string range $::raw_sectors(0,VMGM_MAT) $offset [expr $offset + 7]]
		set txt "$txt\n- Audio $s:  [audio_description $attr true]"
	}
	if {$numsubpics > 0} {
		set attr [string range $::raw_sectors(0,VMGM_MAT) 0X156 0X15B]
		set txt "$txt\n- Subpic:   [subpic_description $attr]"
	}

	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		set numaudios [get_int ::raw_sectors($vts,VTSI_MAT) 0X102 2]
		set numsubpics [get_int ::raw_sectors($vts,VTSI_MAT) 0X154 2]

		set txt "$txt\n\n"
		set txt "$txt\nVTS $vts Menu streams attributes (VTS_[format %02X $vts]_0.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
		set txt "$txt\n- Video:    [video_description [get_int ::raw_sectors($vts,VTSI_MAT) 0X100 2] true]"
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr 0X104 + ( $s * 8 )]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr $offset + 7]]
			set txt "$txt\n- Audio $s:  [audio_description $attr true]"
		}
		if {$numsubpics > 0} {
			set attr [string range $::raw_sectors($vts,VTSI_MAT) 0X156 0X15B]
			set txt "$txt\n- Subpic:   [subpic_description $attr]"
		}

		set numaudios [get_int ::raw_sectors($vts,VTSI_MAT) 0X202 2]
		set numsubpics [get_int ::raw_sectors($vts,VTSI_MAT) 0X254 2]
		set txt "$txt\n"
		set txt "$txt\nVTS $vts VOB streams attributes (VTS_[format %02X $vts]_0.IFO): $numaudios Audio, $numsubpics Sub-picture streams."
		set txt "$txt\n- Video:    [video_description [get_int ::raw_sectors($vts,VTSI_MAT) 0X200 2] true]"
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr 0X204 + ( $s * 8 )]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr $offset + 7]]
			set txt "$txt\n- Audio $s:  [audio_description $attr true]"
		}
		for {set s 0} {$s < $numsubpics} {incr s} {
			set offset [expr 0X256 + ( $s * 6 )]
			set attr [string range $::raw_sectors($vts,VTSI_MAT) $offset [expr $offset + 5]]
			set txt "$txt\n- Subpic $s: [subpic_description $attr]"
		}
	}
	::main::showtext .streamsinfo "PgcEdit: Streams information" $txt
	return $txt
}


proc set_streams_attribs {parentwin} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}

	set pgctype [lindex $::editor_cmds_from 0]
	set vtsnum  [lindex [split [lindex $::editor_cmds_from 1] ","] 0]
	if {$vtsnum == 0} {
		if {$pgctype == "pgc"} {
			bell
			return
		}
		set vtsname "VMGM"
		set mat_table $::raw_sectors(0,VMGM_MAT)
	} else {
		if {$pgctype == "menu"} {
			set vtsname "VTSM $vtsnum"
		} else {
			set vtsname "VTST $vtsnum"
		}
		set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
	}
	if {$pgctype == "menu"} {
		set astart [expr 0X100]
		set numaudios  [get_int mat_table [expr $astart+2] 2]
		set numsubpics [get_int mat_table [expr $astart+0X54] 2]
	} else {
		set astart [expr 0X200]
		if {$vtsnum > 0} {
			set numaudios  [get_int mat_table [expr $astart+2] 2]
			set numsubpics [get_int mat_table [expr $astart+0X54] 2]
		} else {
			set numaudios  0
			set numsubpics 0
		}
	}

	set w ".streams"
	if {[winfo exists $w]} {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm iconname $w Dialog
	wm protocol $w WM_DELETE_WINDOW {set ::streams::done 0}
	bind $w <Escape> "set ::streams::done 0"
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {catch {wm attributes $w -toolwindow 1}}
	wm title $w "Streams attributes"

	if {$pgctype == "pgc"} {set domain "Title"} else {set domain "Menu"}
	label  $w.lab -text "$vtsname $domain domain streams:" -anchor w
	pack   $w.lab -side top -fill x -padx 4

	# VIDEO

	set vfmt [get_int mat_table $astart 2]
	set ::streams::screenformat  [expr $vfmt & 0X0C00]
	set ::streams::autoletterbox [expr $vfmt & 0x0100]
	set ::streams::autopanscan   [expr $vfmt & 0x0200]
	set ::streams::lbcropped     [expr $vfmt & 0x0002]

	frame  $w.f1 -borderwidth 2 -relief groove
	pack   $w.f1 -padx 4 -pady 2 -fill x
	label  $w.f1.lab -text "Video:" -width 9 -anchor w
	pack   $w.f1.lab -side left
	frame  $w.f1.af
	pack   $w.f1.af -side left
	radiobutton $w.f1.af.r1 -text "4:3"  -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::screenformat -value 0
	pack   $w.f1.af.r1 -fill x
	radiobutton $w.f1.af.r2 -text "16:9" -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::screenformat -value [expr 0X0C00]
	pack   $w.f1.af.r2 -fill x

	frame  $w.f1.at
	pack   $w.f1.at -side left
	checkbutton $w.f1.at.r2 -text "Automatic Pan&Scan"  -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::autopanscan   -offvalue [expr 0X0200] -onvalue 0
	pack   $w.f1.at.r2 -fill x
	checkbutton $w.f1.at.r1 -text "Automatic Letterbox" -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::autoletterbox -offvalue [expr 0X0100] -onvalue 0
	pack   $w.f1.at.r1 -fill x
	checkbutton $w.f1.at.r3 -text "Letterbox: Top & bottom cropped" -anchor w -pady 0 -bd $::config(min_bd) -highlightthickness 1 -variable ::streams::lbcropped -offvalue 0 -onvalue [expr 0X0002]
	pack   $w.f1.at.r3 -fill x

	label  $w.f1.d -text "[video_description $vfmt false]" -justify left -anchor w
	pack   $w.f1.d -side left

	# AUDIO

	frame  $w.f2 -borderwidth 2 -relief groove
	pack   $w.f2 -padx 4 -pady 2 -ipady 1 -fill x
	if {$numaudios == 0} {
		label $w.f2.lab -text "No Audio streams"
		pack  $w.f2.lab
	} else {
		frame  $w.f2.pad -borderwidth 1
		pack   $w.f2.pad
	}
	for {set s 0} {$s < $numaudios} {incr s} {
		set offset [expr $astart+4+($s*8)]
		set aattr [string range $mat_table $offset [expr $offset + 7]]
		set attrlist [audio_attributes $aattr]
		set ::streams::audio_lang_code_$s [lindex $attrlist 0]
		set tmp ::streams::audio_lang_code_$s
		if {[subst $$tmp] == "--"} {set ::streams::audio_lang_code_$s ""}
		set ::streams::audio_code_ext_$s [lindex $attrlist 1]

		frame  $w.f2.f$s
		pack   $w.f2.f$s -fill x
		label  $w.f2.f$s.lab -text "Audio $s:" -width 9 -anchor w
		pack   $w.f2.f$s.lab -side left
		entry  $w.f2.f$s.lc -textvariable ::streams::audio_lang_code_$s -width 2 \
				 -font $::config(fixed_width_font) -bd 1 \
				 -validate all -validatecommand "::streams::validate_langcode %P"
		pack   $w.f2.f$s.lc -side left -padx 2
		menubutton $w.f2.f$s.ce -textvariable ::streams::audio_code_ext_$s \
				-pady 0 -width 22 -menu $w.f2.f$s.ce.audiocodeext
		pack   $w.f2.f$s.ce -side left -padx 2
		menu $w.f2.f$s.ce.audiocodeext -tearoff 0
		foreach i $::streams::audio_code_ext_list {
			$w.f2.f$s.ce.audiocodeext add command -label $i \
					-command "set ::streams::audio_code_ext_$s \"$i\""
		}
		label  $w.f2.f$s.d -text [audio_description $aattr false]
		pack   $w.f2.f$s.d -side left
	}

	# SUBPIC

	frame  $w.f3 -borderwidth 2 -relief groove
	pack   $w.f3 -padx 4 -pady 2 -ipady 1 -fill x
	set break 16
	if {$numsubpics == 0} {
		label $w.f3.lab -text "No Sub-Picture streams"
		pack  $w.f3.lab
	} else {
		frame  $w.f3.pad -borderwidth 1
		pack   $w.f3.pad
		frame  $w.f3.f1
		pack   $w.f3.f1 -side left
		frame  $w.f3.f2
		pack   $w.f3.f2 -side left -fill y
		if {$numsubpics > 16} {
			set break [expr ($numsubpics+1)/2]
		}
	}
	for {set s 0} {$s < $numsubpics} {incr s} {
		set offset [expr $astart+0X56+($s*6)]
		set sattr [string range $mat_table $offset [expr $offset + 5]]
		set attrlist [subpic_attributes $sattr]
		set ::streams::subpic_lang_code_$s [lindex $attrlist 0]
		set tmp ::streams::subpic_lang_code_$s
		if {[subst $$tmp] == "--"} {set ::streams::subpic_lang_code_$s ""}
		set ::streams::subpic_code_ext_$s [lindex $attrlist 1]

		set fr $w.f3.f1
		if {$s >= $break} {set fr $w.f3.f2}
		frame  $fr.f$s
		pack   $fr.f$s -fill x
		label  $fr.f$s.lab -text "SubPic $s:" -width 9 -anchor w
		pack   $fr.f$s.lab -side left
		entry  $fr.f$s.lc -textvariable ::streams::subpic_lang_code_$s -width 2 \
				 -font $::config(fixed_width_font) -bd 1 \
				 -validate all -validatecommand "::streams::validate_langcode %P"
		pack   $fr.f$s.lc -side left -padx 2
		menubutton $fr.f$s.ce -textvariable ::streams::subpic_code_ext_$s \
				-pady 0 -width 22 -menu $fr.f$s.ce.subpiccodeext
		pack   $fr.f$s.ce -side left -padx 2
		menu $fr.f$s.ce.subpiccodeext -tearoff 0
		foreach i $::streams::subpic_code_ext_list {
			if {$i != "-"} {
				$fr.f$s.ce.subpiccodeext add command -label $i \
						-command "set ::streams::subpic_code_ext_$s \"$i\""
			}
		}
	}

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x
	button $w.fb.lc -text " ISO639 language codes " -command "::main::showtextfile .langcodes \
			{ISO639 language codes} {$::config(apppath)/lib/data/lang_codes.txt} grab"
	pack   $w.fb.lc -side left -padx 3
	frame  $w.fb.pad
	pack   $w.fb.pad -expand 1 -fill x -side left
	button $w.fb.ok -text "OK" -width 10 -command {set ::streams::done 1}
	pack   $w.fb.ok -side left -padx 3
	button $w.fb.cancel -text "Cancel" -width 10 -command {set ::streams::done 0}
	pack   $w.fb.cancel -side left -padx 2

	# window geometry
	update
	set ww [winfo reqwidth $w]
	set wh [winfo reqheight $w]
	set wx [expr [winfo rootx .]+(([winfo width  .]-$ww)/2)]
	set wy [expr [winfo rooty .]+(([winfo height .]-$wh)/2)-25]
	if {$wx < 10} {set wx 10}
	if {$wy < 10} {set wy 10}
	wm geometry $w ${ww}x$wh+$wx+$wy
	wm resizable $w 0 1

	wm deiconify $w
	grab set $w

	set cont true
	set ::streams::done -1
	while {$cont==true} {
		set cont false
		tkwait variable ::streams::done
		if {$::streams::done} {
			for {set s 0} {$s < $numaudios} {incr s} {
				set tmp ::streams::audio_lang_code_$s
				if {[string length [subst $$tmp]]==1} {
					focus $w.f2.f$s.lc
					bell
					set cont true
					break
				}
			}
			if {! $cont} {
				for {set s 0} {$s < $numsubpics} {incr s} {
					set tmp ::streams::subpic_lang_code_$s
					if {[string length [subst $$tmp]]==1} {
						set ent $w.f3.f1.f$s.lc
						if {$s >= $break} {set ent $w.f3.f2.f$s.lc}
						focus $ent
						bell
						set cont true
						break
					}
				}
			}
		}
	}
	grab release $w
	if {$parentwin != "."} {grab set $parentwin}
	destroy $w
	if $::streams::done {
		::main::dvd_modified true false

		# VIDEO
		set oldval [expr [get_int mat_table $astart 2] & 0XF0FD]
		set_int mat_table $astart [expr $oldval | $::streams::screenformat | \
				$::streams::autoletterbox | $::streams::autopanscan | \
				$::streams::lbcropped] 2

		# AUDIO
		for {set s 0} {$s < $numaudios} {incr s} {
			set offset [expr $astart+4+($s*8)]
			set aattr [string range $mat_table $offset [expr $offset + 7]]
			set byte0 [get_int aattr 0 1]
			set byte5 [get_int aattr 5 1]

			set oldval [expr $byte0 & 0XF3]
			set tmp ::streams::audio_lang_code_$s
			set lc [subst $$tmp]
			if {[string length $lc]==2} {
				set_int aattr 0 [expr $oldval+4] 1
				set aattr "[string range $aattr 0 1]$lc[string range $aattr 4 7]"
			} else {
				if {[string length $lc]!=0} bell
				set_int aattr 0 $oldval 1
			}

			set tmp ::streams::audio_code_ext_$s
			set ce [subst $$tmp]
			if {$ce == $::streams::illegalextlabel} {
				set_int aattr 5 0 1
			} else {
				set_int aattr 5 [lsearch $::streams::audio_code_ext_list $ce] 1
			}

			set mat_table "[string range $mat_table 0 [expr $offset-1]]$aattr[string range $mat_table [expr $offset+8] end]"
		}

		# SUBPIC
		for {set s 0} {$s < $numsubpics} {incr s} {
			set offset [expr $astart+0X56+($s*6)]
			set sattr [string range $mat_table $offset [expr $offset + 5]]
			set byte0 [get_int sattr 0 1]
			set oldval [expr $byte0 & 0XFC]
			set tmp ::streams::subpic_lang_code_$s
			set lc [subst $$tmp]
			if {[string length $lc]==2} {
				set_int sattr 0 [expr $oldval+1] 1
				set sattr "[string range $sattr 0 1]$lc[string range $sattr 4 5]"
			} else {
				if {[string length $lc]!=0} bell
				set_int sattr 0 $oldval 1
			}

			set tmp ::streams::subpic_code_ext_$s
			set ce [subst $$tmp]
			if {$ce == $::streams::illegalextlabel || $ce == "-"} {
				set_int sattr 5 0 1
			} else {
				set_int sattr 5 [lsearch $::streams::subpic_code_ext_list $ce] 1
			}

			set mat_table "[string range $mat_table 0 [expr $offset-1]]$sattr[string range $mat_table [expr $offset+6] end]"
		}

		if {$vtsnum == 0} {
			set ::raw_sectors(0,VMGM_MAT) $mat_table
		} else {
			set ::raw_sectors($vtsnum,VTSI_MAT) $mat_table
		}
	}
}


proc video_attributes {vattr} {

	if {[expr $vattr & 0X4000] == 0} {
		set coding_mode "MPEG-1"
	} else {
		set coding_mode "MPEG-2"
	}

	if {[expr $vattr & 0X1000] == 0} {
		set standard "NTSC"
	} else {
		set standard "PAL"
	}

	if {[expr $vattr & 0X0C00] == 0} {
		set aspect "4:3"
	} else {
		set aspect "16:9"
	}

	if {[expr $vattr & 0X0010] == 0} {
		set bitrate "VBR"
	} else {
		set bitrate "CBR"
	}

	set m [expr ($vattr & 0X0300) / 0X100]
	switch $m {
		3 {set auto_mode ""}
		2 {set auto_mode "Automatic Letterbox"}
		1 {set auto_mode "Automatic Pan&Scan"}
		0 {set auto_mode "Automatic Letterbox & Pan&Scan"}
		default {set auto_mode "Unknow Automatic mode #$m"}
	}

	set r [expr ($vattr & 0X000C) / 4]
	if {$standard == "NTSC"} {
		set h 480
	} else {
		set h 576
	}
	switch $r {
		0 {set w 720}
		1 {set w 704}
		2 {set w 352}
		3 {set w 352; set h [expr $h / 2]}
	}
	set resolution ${w}x${h}

	if {[expr $vattr & 0X0002] == 0} {
		set letterboxed "not cropped"
	} else {
		set letterboxed "top & bottom cropped"
	}

	return [list $aspect $standard $resolution $letterboxed $coding_mode $bitrate $auto_mode]
}
proc video_description {vattr showall} {
	set lst [video_attributes $vattr]
	if $showall {
		set sep ", "
	} else {
		set sep "\n"
	}
	if $showall {
		set txt "[lindex $lst 0] "
	} else {
		set txt ""
	}
	set txt "${txt}[lindex $lst 1] ([lindex $lst 2])${sep}[lindex $lst 4] ([lindex $lst 5])"

	if $showall {
		set auto_mode [lindex $lst 6]
		if {$auto_mode != ""} {
			set txt "$txt, $auto_mode"
			set txt "$txt (letterbox: [lindex $lst 3])"
		}
	}
	return $txt
}


proc audio_attributes {aattr} {

	set byte0 [get_int aattr 0 1]

	set c [expr ($byte0 & 0XE0) / 0X20]
	switch $c {
		0 {set coding_mode "AC3"}
		2 {set coding_mode "MPEG-1"}
		3 {set coding_mode "MPEG-2"}
		4 {set coding_mode "LPCM"}
		6 {set coding_mode "DTS"}
		default {set coding_mode "Unknow coding mode #$c"}
	}

	if {[expr $byte0 & 0X0C] == 0} {
		set lang_code "--"
	} else {
		set lang_code [string range $aattr 2 3]
	}

	set a [expr $byte0 & 0X03]
	switch $a {
		0 {set applic_mode "Unspecified application mode"}
		1 {set applic_mode "Karaoke"}
		2 {
			set byte7 [get_int aattr 7 1]
			if {[expr $byte7 & 8] == 8} {
				set applic_mode "Dolby-Surround"
			} else {
				set applic_mode "Surround"
			}
		}
		3 {set applic_mode "Unknow applic. mode #$a"}
	}

	set byte1 [get_int aattr 1 1]

	set num_channels [expr ($byte1 & 7) + 1]

	set byte5 [get_int aattr 5 1]
	if {$byte5 < 5} {
		set code_ext [lindex $::streams::audio_code_ext_list $byte5]
	} else {
		set code_ext $::streams::illegalextlabel
	}

	return [list $lang_code $code_ext $coding_mode $num_channels $applic_mode]
}
proc audio_description {vattr showall} {
	set lst [audio_attributes $vattr]
	if $showall {
		set txt "[lindex $lst 0], [lindex $lst 1], "
	} else {
		set txt ""
	}
	set txt "${txt}[lindex $lst 2] ([lindex $lst 3] channel"
	if {[lindex $lst 3] > 1} {
		set txt "${txt}s)"
	} else {
		set txt "${txt})"
	}
	if $showall {
		set txt "$txt, [lindex $lst 4]"
	}
	return $txt
}


proc subpic_attributes {spattr} {

	set byte0 [get_int spattr 0 1]

	if {[expr $byte0 & 0X03] == 0} {
		set lang_code "--"
	} else {
		set lang_code [string range $spattr 2 3]
	}

	set byte5 [get_int spattr 5 1]
	if {$byte5 < 16} {
		set code_ext [lindex $::streams::subpic_code_ext_list $byte5]
		if {$code_ext == "-"} {
			set code_ext $::streams::illegalextlabel
		}
	} else {
		set code_ext $::streams::illegalextlabel
	}

	return [list $lang_code $code_ext]
}
proc subpic_description {vattr} {
	set lst [subpic_attributes $vattr]
	return "[lindex $lst 0], [lindex $lst 1]"
}

set valid_lang_codes {
	aa ab af am ar as ay az ba be bg bh bi bn bo br ca co cs cy da de dz el en eo es et
	eu fa fi fj fo fr fy ga gd gl gn gu ha hi hr hu hy ia ie ik in is it iw ja ji jw
	ka kk kl km kn ko ks ku ky la ln lo lt lv mg mi mk ml mn mo mr ms mt my na ne nl
	no oc om or pa pl ps pt qu rm rn ro ru rw sa sd sg sh si sk sl sm sn so sq sr ss
	st su sv sw ta te tg th ti tk tl tn to tr ts tt tw uk ur uz vi vo wo xh yo zh zu
}

proc validate_langcode {newcode} {
	if {[string length $newcode]==""} {
		return true
	}
	set r [valid_langcode $newcode]
	if {! $r} bell
	return $r
}

proc valid_langcode {newcode} {
	if {[string length $newcode]>2} {
		return false
	}
	if {! [string is alpha $newcode]} {
		return false
	}
	if {! [string is lower $newcode]} {
		return false
	}
	if {[string length $newcode]==2 && [lsearch $::streams::valid_lang_codes $newcode] < 0} {
		return false
	}
	return true
}


####### main proc
proc get_pgc_audio_descr {pgctype pgcindex streamnum} {
	set streamnum [expr $streamnum & 0x0F]
	if {$streamnum == 15} {return "15: None"}
	set vts [lindex [split $pgcindex ,] 0]
	if {$vts == 0} {
		if {$pgctype == "pgc"} {
			return ""
		} else {
			set data $::raw_sectors(0,VMGM_MAT)
			set numaudios [get_int data 0X102 2]
			set offset [expr 0X104 + ( $streamnum * 8 )]
		}
	} else {
		if {$pgctype == "menu"} {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numaudios [get_int data 0X102 2]
			set offset [expr 0X104 + ( $streamnum * 8 )]
		} else {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numaudios [get_int data 0X202 2]
			set offset [expr 0X204 + ( $streamnum * 8 )]
		}
	}
	set attr [string range $data $offset [expr $offset + 7]]

	set lst [audio_attributes $attr]
	return "[expr $streamnum & 0x07]: [lindex $lst 0], [lindex $lst 1] ([lindex $lst 2])"
}
proc get_pgc_subpic_descr {pgctype pgcindex streamnum} {
	set streamnum [expr $streamnum & 0x7F]
	if {$streamnum == 62} {return "62: None"}
	if {$streamnum == 63} {return "63: Forced"}
	if {[expr $streamnum & 0x40] == 0} {
		set mode "Off"
	} else {
		set mode "On"
	}
	set streamnum [expr $streamnum & 0x1F]

	set vts [lindex [split $pgcindex ,] 0]
	if {$vts == 0} {
		if {$pgctype == "pgc"} {
			return ""
		} else {
			set data $::raw_sectors(0,VMGM_MAT)
			set numsubpics [get_int data 0X154 2]
			set offset [expr 0X156 + ( $streamnum * 6 )]
		}
	} else {
		if {$pgctype == "menu"} {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numsubpics [get_int data 0X154 2]
			set offset [expr 0X156 + ( $streamnum * 6 )]
		} else {
			set data $::raw_sectors($vts,VTSI_MAT)
			set numsubpics [get_int data 0X254 2]
			set offset [expr 0X256 + ( $streamnum * 6 )]
		}
	}
	set attr [string range $data $offset [expr $offset + 5]]
	return "$streamnum $mode: [subpic_description $attr]"
}

}	;# end namespace streams
 